# -*- coding: utf-8 -*-

# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2019 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENSE.TXT
# show cagefs usage statistics - work for cloudlinux-config utility

import os
import subprocess
from typing import Optional  # NOQA

BIN_EXEC = '/usr/sbin/cagefsctl'
CAGEFS_STATUS_NOT_INITIALIZED = 'Not initialized'
CAGEFS_STATUS_NOT_INSTALLED = 'Not installed'


def get_cagefs_user_mode():
    # type: () -> Optional[str]
    """
    Get Cagefs user mode, execute cagefsctl --display-user-mode cmd
    :return string -> cagefs user mode
    """

    if os.path.exists(BIN_EXEC):
        try:
            with subprocess.Popen(
                [BIN_EXEC, '--display-user-mode'],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
            ) as proc:
                out, _ = proc.communicate()
            try:
                return out.split(':')[1].strip()
            except IndexError:
                return None
        except (OSError, IOError):
            return None


def get_cagefs_status():

    """
    Get Cagefs user status, execute cagefsctl --cagefs-status cmd
    :return string -> cagefs status
    """

    if not os.path.exists(BIN_EXEC):
        return CAGEFS_STATUS_NOT_INSTALLED
    if not check_cagefs_initialized():
        return CAGEFS_STATUS_NOT_INITIALIZED
    try:
        with subprocess.Popen(
            [BIN_EXEC, '--cagefs-status'],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        ) as proc:
            out, _ = proc.communicate()
        return out.strip()
    except (OSError, IOError):
        return None


def get_quantity(enabled):

    """
    Get Cagefs quantity if enabled/disabled users, execute
    cagefsctl --list-enabled cmd for enabled users
    execute cagefsctl --list-disabled cmd for disabled users
    :param enabled: true/false flag for quantity enabled/disabled users
    :return int -> quantity of users
    """

    if os.path.exists(BIN_EXEC):
        try:
            if enabled:
                cmd = '--list-enabled'
            else:
                cmd = '--list-disabled'
            # do not need to check returncode for cagefsctl, not always returns 0
            with subprocess.Popen(
                [BIN_EXEC, cmd],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
            ) as proc:
                out, _ = proc.communicate()
            try:
                stat = list(out.split('\n'))
                if not stat[0]:
                    return 0
                return int(stat[0].split(' ')[0])
            except (IndexError, ValueError):
                return None
        except (OSError, IOError):
            return None


def check_cagefs_initialized():
    """
    Check if cagefs is initialized with help of --check-cagefs-skeleton command
    --check-cagefs-skeleton returns rc 0 only if it is initialized
    """
    try:
        subprocess.check_output([BIN_EXEC, '--check-cagefs-initialized'],
                                stderr=subprocess.STDOUT, shell=False)
        return True
    except (subprocess.CalledProcessError, OSError):
        return False
