# -*- coding: utf-8 -*-

# config.py - work code for cloudlinux-config utility
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2019 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENSE.TXT

import argparse
import json
import sys
import time
from abc import ABC, abstractmethod

from clcommon import cpapi
from clcommon.clexception import FormattedException
from clcommon.const import Feature
from clcommon.cpapi import is_admin, is_panel_feature_supported
from clcommon.utils import is_litespeed_running
from cli_utils import print_dictionary, replace_params
from cllicense import CloudlinuxLicenseLib
from cllimits.clquota_lib import ClQuotaLib
from lvectllib import is_active_reseller_limits, lve

from . import db_governor_lib, ui_config_lib
from .cl_config_arg_parse import parse_cloudlinux_config_opts
from .clconfig_utils import (
    boolean_to_0_1,
    boolean_to_on_off,
    boolean_to_yes_no,
    is_int,
    is_positive_int,
    is_string,
    is_string_or_none,
    time_unit_to_letter,
)
from .db_info_lib import MysqlInfo


class ValidateDataException(FormattedException):
    pass


class ClConfingBase(ABC):

    def __init__(self):
        self._is_json = False
        self._opts = {}

    @abstractmethod
    def _get_data_list(self):
        """
        Return list of pairs (input_dict_keys, validator_function).
        """
        pass

    def _print_data(self, data, force_json=False, result="success"):
        """
        Output data wrapper
        :param: `dict` data - data for output to stdout
        :param: `bool` force_json - always output json format
        """
        if isinstance(data, dict):
            data.update({"result": result, "timestamp": time.time()})
        print_dictionary(data, self._is_json or force_json)

    def _error_and_exit(self, message, error_code=1):
        """
        Print error and exit
        :param message: Dictionary with keys "result" as string and optional "context" as dict
        """
        if self._is_json:
            message.update({"timestamp": time.time()})
            print_dictionary(message, True)
        else:
            try:
                print(str(message["result"]) % message.get("context", {}))
            except KeyError as e:
                print(f"Error: {e} [{message}]")
        sys.exit(error_code)

    def _validate_parameters_single(self, data_dict, reseller=None):
        """
        Validate config single parameters.
        """
        # Control data list struct:
        #  (input_dictionary_keys, validator_function, key_name_for_exception)
        control_data_list = self._get_data_list()
        for control_data_item in control_data_list:
            try:
                input_dict_keys, validator_function = control_data_item
                # ['faultsNotification']['minimumNumberOfFaultsToNotify']['reseller']
                # goes to 'faultsNotification/minimumNumberOfFaultsToNotify/reseller'
                key_name_for_exception = input_dict_keys\
                    .replace('[', '')\
                    .replace('\'', '')\
                    .replace(']', '/')\
                    .rstrip('/')
                input_value = eval(f"data_dict{input_dict_keys}")
                if validator_function(input_value) is None:
                    raise ValidateDataException({'message': "Value error in JSON key(s) options/%(key)s",
                                                 'context': {'key': key_name_for_exception}})
            except KeyError:
                pass


class ClConfigWithoutLve(ClConfingBase):
    """
    Perform cloudlinux-config utility for CloudLinux OS Solo.
    """

    def run(self, args):
        licence = CloudlinuxLicenseLib()
        if not licence.get_license_status():
            self._is_json = True
            return self._error_and_exit({"result": "Cloudlinux license isn't valid"})
        # get arguments
        self._is_json = '--json' in args
        opts = self._parse_args(args)
        self._opts = vars(opts)  # Convert parsed options to dictionary
        if self._opts.get('command') == 'set':
            self._set_run()
        else:
            self._get_run()

    def _set_run(self):
        """
        Set parameters from self._opts
        :return:
        """
        data_dict = {}
        try:
            data_dict = json.loads(self._opts['data'])['options']
        except (ValueError, KeyError):
            # Any exception - invalid JSON passed
            self._error_and_exit({'result': "ERROR: Invalid JSON string passed in --data key"})
        # Validate data_dict
        try:
            self._validate_input_dictionary(data_dict)
        except ValidateDataException as e:
            # if error - print it and exit
            self._error_and_exit({"result": f"Error: {e}"})
        result = {'result': '', 'timestamp': time.time(), 'context': {}}
        ##################################################################
        # Set UI parameters
        ui_parameters = {}
        if 'uiSettings' in data_dict:
            ui_parameters['uiSettings'] = data_dict['uiSettings']
        try:
            ui_config_lib.set_ui_config(ui_parameters)
        except ui_config_lib.UIConfigException as e:
            result['result'] += e.message + "; "
            result['context'].update(e.context)
        ###############################
        # Print success result
        if not result['result']:
            result['result'] = 'success'
        if len(result['context']) == 0:
            del result['context']
        print_dictionary(result, True)

    def _get_run(self):
        """
        Run 'get' command
        """
        result = {'result': 'success', 'timestamp': time.time()}
        full_config = {}
        full_config.update(ui_config_lib.get_ui_config())
        result['options'] = full_config

        result['supported_cl_features'] = [
            component for component, is_supported in
            cpapi.get_supported_cl_features().items()
            if is_supported]
        supported_cl_features_list = result['supported_cl_features']

        if is_litespeed_running() and Feature.LSAPI in supported_cl_features_list:
            supported_cl_features_list.remove(Feature.LSAPI)
        print_dictionary(result, self._is_json)

    def _parse_args(self, argv):
        """
        Parse CLI arguments
        """
        parser = argparse.ArgumentParser(
            prog='cloudlinux-config',
            description='Utility to get/set Cloudlinux UI config',
        )
        subparsers = parser.add_subparsers(
            title='commands', dest='command')
        # Get subparser
        subparsers.add_parser('get', help='get UI config')
        # Set subparser
        set_subparser = subparsers.add_parser('set', help='set UI config')
        set_subparser.add_argument('--data', required=True, help='data to set')
        opts, _ = parser.parse_known_args(argv)
        return opts

    def _validate_input_dictionary(self, data_dict):
        """
        Check data validity in data_dict.
        For CL OS Solo we check only single parameters.
        :param data_dict: Dictionary with input data
        :return: None, throws ValidateDataException if any error detected
        """
        self._validate_parameters_single(data_dict)

    def _get_data_list(self):
        # Only UI settings for CL OS Solo
        return [
            ("['uiSettings']['hideRubyApp']", boolean_to_0_1),
            ("['uiSettings']['hidePythonApp']", boolean_to_0_1),
            ("['uiSettings']['hideNodeJsApp']", boolean_to_0_1),
            ("['uiSettings']['hidePHPextensions']", boolean_to_0_1),
            ("['uiSettings']['hideLVEUserStat']", boolean_to_0_1),
            ("['uiSettings']['hideXrayApp']", boolean_to_0_1),
            ("['uiSettings']['hideAccelerateWPApp']", boolean_to_0_1),
        ]


class ClConfig(ClConfingBase):
    """
    Perform cloudlinux-config utility for CloudLinux OS Shared and Shared Pro.
    """

    def run(self, args):
        licence = CloudlinuxLicenseLib()
        if not licence.get_license_status():
            self._is_json = True
            return self._error_and_exit({"result": "Cloudlinux license isn't valid"})
        self._is_json = "--json" in args
        # get arguments
        self._opts = self._parse_args(args)
        if self._opts['--for-reseller'] and not lve.reseller_limit_supported():
            self._error_and_exit({
                'result': 'ERROR: reseller limits are not supported, can not use --for-reseller option.',
            })
        if self._opts['get']:
            # Get mode
            self._get_run()
        else:
            # Set mode
            self._set_run()

    def _set_run(self):
        """
        Set parameters from self._opts
        :return:
        """
        reseller = self._opts.get('--for-reseller', None)
        if reseller is not None and \
           not is_admin(reseller) and \
           not is_active_reseller_limits(reseller):
            self._error_and_exit({
                'result': 'ERROR: reseller limits for %(reseller)s are disabled.',
                'context': {'reseller': reseller}
            })

        if self._opts['--reset-inodes-limit']:
            if reseller is not None:
                self._error_and_exit({
                    'result': "ERROR: cannot reset inode limits for reseller user"
                })

            # Reset inodes limits
            cl_quota = ClQuotaLib()
            if not cl_quota.is_clquota_activated():
                result = "ERROR: quota isn't activated"
            else:
                result = 'success'
                cl_quota.reset_inodes_limits()
            # Print success result
            print_dictionary({'result': result, 'timestamp': time.time()}, True)
            return
        if not self._opts['--data']:
            # no data to set
            self._error_and_exit({'result': "ERROR: You must provide parameters to set"})
        data_dict = {}
        try:
            data_dict = json.loads(self._opts['--data'])['options']
        except Exception:
            # Any exception - invalid JSON passed
            self._error_and_exit({'result': "ERROR: Invalid JSON string passed in --data key"})
        # Validate data_dict
        try:
            self._validate_input_dictionary(data_dict, reseller=reseller)
        except ValidateDataException as e:
            # if error - print it and exit
            self._error_and_exit({"result": f"Error: {e}"})
        result = {'result': '', 'timestamp': time.time(), 'context': {}}
        # to do not copy-paste strings with camel cases
        (inode_limits, ui_settings, faults_notification, mysql_gov) = (
            'inodeLimits', 'uiSettings', 'faultsNotification', 'mySQLGovSettings'
        )
        ##################################################################
        # 0. Check that reseller doesn't change something except faultsNotification
        if reseller is not None:
            from . import lve_stats2_reseller_lib as lve_stats2_notif  # pylint: disable=import-outside-toplevel
            admin_only_settings = {inode_limits, ui_settings, mysql_gov}
            # check that data_dict doesn't contain these keys
            if len(admin_only_settings.intersection(set(data_dict))) != 0:
                self._error_and_exit({
                    'result': f"Error: can only configure {faults_notification} for reseller"
                })
        else:
            from . import lve_stats2_lib as lve_stats2_notif  # pylint: disable=import-outside-toplevel
        ###############################
        # 1. Set UI parameters
        ui_parameters = {}
        if inode_limits in data_dict:
            ui_parameters[inode_limits] = data_dict[inode_limits]
        if ui_settings in data_dict:
            ui_parameters[ui_settings] = data_dict[ui_settings]
        try:
            ui_config_lib.set_ui_config(ui_parameters)
        except ui_config_lib.UIConfigException as e:
            result['result'] += e.message + "; "
            result['context'].update(e.context)
        ###############################
        # 2. Set lve-stats2 parameters
        if faults_notification in data_dict:
            try:
                lve_stats2_notif.set_notification(data_dict[faults_notification],
                                                  reseller=reseller)
            except lve_stats2_notif.LveStats2Exception as e:
                result['result'] += e.message + "; "
                result['context'].update(e.context)
        ###############################
        # 3. Set MySQL Governor parameters
        if mysql_gov in data_dict:
            try:
                db_governor_lib.set_db_gov_config(data_dict[mysql_gov])
            except db_governor_lib.DBGovernorException as e:
                result['result'] += e.message + "; "
                result['context'].update(e.context)
        # Print success result
        if not result['result']:
            result['result'] = 'success'
        if len(result['context']) == 0:
            del result['context']
        print_dictionary(result, True)

    def _get_run(self):
        """
        Run 'get' command
        """
        result = {'result': 'success', 'timestamp': time.time()}
        full_config = {}
        reseller = None
        # pylint: disable=import-outside-toplevel
        if self._opts['--for-reseller']:
            reseller = self._opts['--for-reseller']
            from . import lve_stats2_reseller_lib as lve_stats2_notif
        else:
            from . import lve_stats2_lib as lve_stats2_notif
        # pylint: enable=import-outside-toplevel

        if reseller is None:
            # 1. Get lve-stats2 config
            if is_panel_feature_supported(Feature.LVE):
                full_config.update(lve_stats2_notif.get_notification())
            # 2. Get UI config
            full_config.update(ui_config_lib.get_ui_config())
            # 3. Get MySQL Governor parameters
            try:
                full_config.update(db_governor_lib.get_db_gov_config())
            except db_governor_lib.DBGovernorException as e:
                result["mySqlGov"] = "error"
                result["errors"] = {}
                result["errors"]["mySqlGov"] = str(e)
            # 4. Get MySQL info
            full_config.update({'dbInfo': MysqlInfo().get()})
        else:
            if is_panel_feature_supported(Feature.LVE):
                full_config.update(lve_stats2_notif.get_notification(reseller=reseller, defaults=True))

        result['options'] = full_config
        result['supported_cl_features'] = [
            component for component, is_supported in
            cpapi.get_supported_cl_features().items()
            if is_supported]
        supported_cl_features_list = result['supported_cl_features']
        if is_litespeed_running() and Feature.LSAPI in supported_cl_features_list:
            supported_cl_features_list.remove(Feature.LSAPI)

        print_dictionary(result, self._is_json)

    def _parse_args(self, argv):
        """
        Parse CLI arguments
        """
        status, data = parse_cloudlinux_config_opts(argv, self._is_json)
        if not status:
            # exit with error if can`t parse CLI arguments
            self._error_and_exit(replace_params(data))
        return data

    def _validate_input_dictionary(self, data_dict, reseller=None):
        """
        Checks data validity in data_dict
        :param data_dict: Dictionary with input data
        :param str reseller: Reseller name to validate to config reseller
        :return: None, throws ValidateDataException if any error detected
        """
        self._validate_parameters_single(data_dict, reseller)
        self._validate_parameters_pairs(data_dict)

    def _get_data_list(self):
        return [
            ##################################
            # 1. UI section
            ("['inodeLimits']['showUserInodesUsage']", boolean_to_0_1),
            ("['uiSettings']['hideRubyApp']", boolean_to_0_1),
            ("['uiSettings']['hidePythonApp']", boolean_to_0_1),
            ("['uiSettings']['hideNodeJsApp']", boolean_to_0_1),
            ("['uiSettings']['hidePHPextensions']", boolean_to_0_1),
            ("['uiSettings']['hideLVEUserStat']", boolean_to_0_1),
            ("['uiSettings']['hideXrayApp']", boolean_to_0_1),
            ("['uiSettings']['hideAccelerateWPApp']", boolean_to_0_1),
            ##################################
            # 2. Checks lve-stats section
            ("['faultsNotification']['faultsToInclude']['cpu']", boolean_to_yes_no),
            ("['faultsNotification']['faultsToInclude']['nproc']", boolean_to_yes_no),
            ("['faultsNotification']['faultsToInclude']['io']", boolean_to_yes_no),
            ("['faultsNotification']['faultsToInclude']['mem']", boolean_to_yes_no),
            ("['faultsNotification']['faultsToInclude']['concurrentConnections']", boolean_to_yes_no),
            ("['faultsNotification']['faultsToInclude']['iops']", boolean_to_yes_no),
            ("['faultsNotification']['notify']['admin']['period']", is_positive_int),
            ("['faultsNotification']['notify']['admin']['unitOfTime']", time_unit_to_letter),
            ("['faultsNotification']['notify']['user']['period']", is_positive_int),
            ("['faultsNotification']['notify']['user']['unitOfTime']", time_unit_to_letter),
            ("['faultsNotification']['notifyAdmin']", boolean_to_yes_no),
            ("['faultsNotification']['minimumNumberOfFaultsToNotify']['admin']", is_int),
            ("['faultsNotification']['minimumNumberOfFaultsToNotify']['user']", is_int),
            ("['faultsNotification']['notifyCustomers']", boolean_to_yes_no),
            ("['faultsNotification']['notifyResellers']", boolean_to_yes_no),
            ("['faultsNotification']['notifyResellerCustomers']", boolean_to_yes_no),
            #################################
            # 2.1. Check email section
            ("['faultsNotification']['email']['notifyFromEmail']", is_string_or_none),
            ("['faultsNotification']['email']['notifySubject']", is_string_or_none),
            ("['faultsNotification']['email']['reportAdminMail']", is_string_or_none),
            ("['faultsNotification']['email']['notifyCharset']", is_string_or_none),
            ##################################
            # 3. Checks lve-stats section for reseller
            # Checks lve-stats section
            ("['faultsNotification']['faultsToInclude']['cpu']", boolean_to_yes_no),
            ("['faultsNotification']['faultsToInclude']['nproc']", boolean_to_yes_no),
            ("['faultsNotification']['faultsToInclude']['io']", boolean_to_yes_no),
            ("['faultsNotification']['faultsToInclude']['mem']", boolean_to_yes_no),
            ("['faultsNotification']['faultsToInclude']['concurrentConnections']", boolean_to_yes_no),
            ("['faultsNotification']['faultsToInclude']['iops']", boolean_to_yes_no),
            ("['faultsNotification']['notify']['reseller']['period']", is_positive_int),
            ("['faultsNotification']['notify']['reseller']['unitOfTime']", time_unit_to_letter),
            ("['faultsNotification']['notify']['customer']['period']", is_positive_int),
            ("['faultsNotification']['notify']['customer']['unitOfTime']", time_unit_to_letter),
            ("['faultsNotification']['notifyResellerOnCustomers']", boolean_to_yes_no),
            ("['faultsNotification']['minimumNumberOfFaultsToNotify']['reseller']", is_int),
            ("['faultsNotification']['minimumNumberOfFaultsToNotify']['customer']", is_int),
            ("['faultsNotification']['notifyReseller']", boolean_to_yes_no),
            ("['faultsNotification']['notifyResellers']", boolean_to_yes_no),
            ("['faultsNotification']['notifyCustomers']", boolean_to_yes_no),
            ##################################
            # 3. Checks MySQL governor section
            ("['mySQLGovSettings']['modeOfOperation']", is_string),
            ("['mySQLGovSettings']['scriptPath']", is_string),
            ("['mySQLGovSettings']['userMaxConnections']", is_int),
            ("['mySQLGovSettings']['logRestrictedUsersQueries']", boolean_to_on_off),
            ("['mySQLGovSettings']['gatherDataForDetailedStats']", boolean_to_on_off),
            ("['mySQLGovSettings']['restrictLog']['logPath']", is_string),
            ("['mySQLGovSettings']['restrictLog']['format']", is_string),
            ("['mySQLGovSettings']['errorLog']['logPath']", is_string),
            ("['mySQLGovSettings']['errorLog']['level']", is_string),
            ("['mySQLGovSettings']['slowQueries']['logPath']", is_string),
            ("['mySQLGovSettings']['slowQueries']['kill']", boolean_to_on_off),
            ("['mySQLGovSettings']['slowQueries']['timeout']", is_int),
            ("['mySQLGovSettings']['restrictType']['mode']", is_string),
            ("['mySQLGovSettings']['unlimit']['unitOfTime']", is_string),
            ("['mySQLGovSettings']['unlimit']['period']", is_int),
            ("['mySQLGovSettings']['restrictedTimePeriods']['level1']['unitOfTime']", is_string),
            ("['mySQLGovSettings']['restrictedTimePeriods']['level1']['period']", is_int),
            ("['mySQLGovSettings']['restrictedTimePeriods']['level2']['unitOfTime']", is_string),
            ("['mySQLGovSettings']['restrictedTimePeriods']['level2']['period']", is_int),
            ("['mySQLGovSettings']['restrictedTimePeriods']['level3']['unitOfTime']", is_string),
            ("['mySQLGovSettings']['restrictedTimePeriods']['level3']['period']", is_int),
            ("['mySQLGovSettings']['restrictedTimePeriods']['level4']['unitOfTime']", is_string),
            ("['mySQLGovSettings']['restrictedTimePeriods']['level4']['period']", is_int),
            ("['mySQLGovSettings']['restrictedTimePeriods']['timeout']['unitOfTime']", is_string),
            ("['mySQLGovSettings']['restrictedTimePeriods']['timeout']['period']", is_int)
        ]

    def _validate_parameters_pairs(self, data_dict):
        """
        Validate config pairs parameters (period and unitOfTime).
        """
        pairs_data_list = [
            ("['faultsNotification']['notify']['admin']['period']",
             "['faultsNotification']['notify']['admin']['unitOfTime']"),
            ("['faultsNotification']['notify']['user']['period']",
             "['faultsNotification']['notify']['user']['unitOfTime']"),
            ("['mySQLGovSettings']['restrictType']['unlimit']['unitOfTime']",
             "['mySQLGovSettings']['restrictType']['unlimit']['period']"),
            ("['mySQLGovSettings']['restrictedTimePeriods']['level1']['unitOfTime']",
             "['mySQLGovSettings']['restrictedTimePeriods']['level1']['period']"),
            ("['mySQLGovSettings']['restrictedTimePeriods']['level2']['unitOfTime']",
             "['mySQLGovSettings']['restrictedTimePeriods']['level2']['period']"),
            ("['mySQLGovSettings']['restrictedTimePeriods']['level3']['unitOfTime']",
             "['mySQLGovSettings']['restrictedTimePeriods']['level3']['period']"),
            ("['mySQLGovSettings']['restrictedTimePeriods']['level4']['unitOfTime']",
             "['mySQLGovSettings']['restrictedTimePeriods']['level4']['period']"),
            ("['mySQLGovSettings']['restrictedTimePeriods']['timeout']['unitOfTime']",
             "['mySQLGovSettings']['restrictedTimePeriods']['timeout']['period']")
        ]
        for pairs_data_item in pairs_data_list:
            keys1, keys2 = pairs_data_item
            val1, val2 = None, None
            try:
                val1 = eval(f"data_dict{keys1}")
            except KeyError:
                pass
            try:
                val2 = eval(f"data_dict{keys2}")
            except KeyError:
                pass
            if (val1 is None and val2 is not None) \
               or (val1 is not None and val2 is None):
                raise ValidateDataException({'message': "Keys 'period' and 'unitOfTime' should be passed both"})
