# coding:utf-8

# license.py - work code for cloudlinux-license utility
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2019 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENSE.TXT

import configparser
import json
import sys
import time
from json import JSONDecodeError
from pathlib import Path
from subprocess import run
from typing import Dict, Optional, Union

import requests
from cli_utils import print_dictionary, replace_params
from cllicense import CloudlinuxLicenseLib

from .cl_license_arg_parse import parse_cloudlinux_packages_opts


def _print_as_json_and_exit(
    error: Optional[str] = None,
    details: Optional[str] = None,
    context: Optional[dict] = None,
) -> None:
    """
    This function makes UI-specific JSON, prints it and exits from the program. It sets "timestamp".
    And if @param error is specified, then field "result" will be filled with this parameter and exit code will be 1.
    Otherwise "result" will be filled with "success" and exit code will be 0.
    @param details: if it's specified, then field "details" will be will be filled with this parameter.
    @param context: if it's specified, then field "context" will be filled with this parameter.
    """

    output: Dict[str, Union[float, str, Dict]]
    output = {"timestamp": time.time()}
    if error is not None:
        output["result"] = error
        exit_code = 1
    else:
        output["result"] = "success"
        exit_code = 0

    if details is not None:
        output["details"] = details
    if context is not None:
        output["context"] = context

    json.dump(output, sys.stdout)
    sys.exit(exit_code)  # TODO exit in deep stacktrace?


def _beautify(validation_errors: dict) -> str:
    """
    This function converts dict of validation errors in a more human-readable presentation as str.
    Ends each line with '\n'.

    @param validation_errors: dict of validation errors
    @return: str
    """
    output = ""
    for field, error in validation_errors.items():
        output += f"- {field}: {error}\n"

    return output


class CloudlinuxLicense:
    """
    Class to check Cloudlinux license
    """

    CONFIG_PATH = Path("/etc/lve-utils/cloudlinux-license.ini")

    def __init__(self):
        self._is_json = False
        self._opts = {}
        self._license_lib = CloudlinuxLicenseLib()

        if self.CONFIG_PATH.exists():
            config = configparser.ConfigParser()
            config.read(self.CONFIG_PATH)
            self.cln_subdomain = "cln-staging" if bool(config["DEFAULT"]["is_stage"]) else "cln"
        else:
            self.cln_subdomain = "cln"

    def run(self, argv):
        """
        Run command action
        """
        self._is_json = "--json" in argv
        # get arguments
        self._opts = self._parse_args(argv)
        if self._opts["set"]:
            self._set_run()
            self._get_run()
        elif self._opts["get_trial"]:
            self._get_trial()
        else:
            self._get_run()
        return 0

    def _get_run(self):
        """
        Run 'get' command
        """
        if self._license_lib.get_license_status():
            # License valid
            license_status = "valid"
        else:
            # License invalid
            license_status = "invalid"
        result_data = {"result": "success", "timestamp": time.time(), "license": license_status}
        print_dictionary(result_data, True)

    def _set_run(self):
        """
        Run 'set' command
        """
        if self._opts["--ip"]:
            # Register server using IP
            self._license_lib.register_and_update_license_by_ip()
        key = self._opts["--key"]
        if key:
            # Register server using key
            self._license_lib.register_and_update_license_by_key(key)

    def _parse_args(self, argv):
        """
        Parse CLI arguments
        """
        status, data = parse_cloudlinux_packages_opts(argv, self._is_json)
        if not status:
            # exit with error if can`t parse CLI arguments
            self._error_and_exit(replace_params(data))
        return data

    def _error_and_exit(self, message, error_code=1):
        """
        Print error and exit
        :param message dict: Dictionary with keys "result" as string and optional "context" as dict
        """
        if self._is_json:
            message.update({"timestamp": time.time()})
            print_dictionary(message, True)
        else:
            try:
                print(str(message["result"]) % message.get("context", {}))
            except KeyError as e:
                print(f"Error: {e} [{message}]")
        sys.exit(error_code)

    def _get_key(self) -> str:
        try:
            config = json.loads(self._opts["--config"])
            edition = config.pop("edition", "CL-Solo")  # Possible values CL,CLS,CLP,CLADMIN (or CL-Solo alias)
        except JSONDecodeError:
            _print_as_json_and_exit("Input config is not valid JSON.", details=self._opts["--config"])

        try:
            response = requests.post(
                f"https://{self.cln_subdomain}.cloudlinux.com/wapi/{edition}/trial",
                json=config,
                timeout=20,
            )
        except requests.RequestException as e:
            _print_as_json_and_exit(
                "We tried to get a trial license, but were unable to connect to the server application. "
                "Please, try again later or contact our support for help.",
                details=str(e),
            )

        try:
            response_json = response.json()
        except JSONDecodeError:
            _print_as_json_and_exit(
                "We tried to get a trial license, but something went wrong on the server side. "
                "Please, try again later or contact our support for help.",
                details=response.text,
            )

        # success
        if "key" in response_json:
            return response_json["key"]

        # errors
        if response_json.get("type") == "validation":
            _print_as_json_and_exit(
                "Input data validation failed: \n %(error_text)s"
                "Please fix form errors and submit again.",
                context={"error_text": _beautify(response_json["errors"])},
            )

        if response_json.get("type") is None:
            _print_as_json_and_exit(response_json["message"])

        # we mustn't be here
        _print_as_json_and_exit(
            "We tried to get a trial license, but something went wrong. "
            "Please, try again later or contact our support for help."
        )
        raise RuntimeError("license key request failed")

    def _get_trial(self) -> None:
        key = self._get_key()

        result = run(
            [
                "clnreg_ks",
                "--force",
                "--activationkey",
                key,
                "--serverUrl",
                f"""https://xmlrpc.{self.cln_subdomain}.cloudlinux.com/XMLRPC""",
            ],
            capture_output=True,
            check=False,
        )
        if result.returncode != 0:
            _print_as_json_and_exit(
                "We created a trial license, but were unable to register the server. Please, try to run this command: "
                f"`{' '.join(result.args)}` a little bit later. Contact our support for help."
            )

        # need this to update license cache file
        result = run(["/usr/bin/cldetect", "--update-license"], capture_output=True, check=False)
        if result.returncode != 0:
            _print_as_json_and_exit(
                "We created a trial license, but were unable to register the server. Please, try to run this command: "
                f"`{' '.join(result.args)}` a little bit later. Contact our support for help."
            )
        else:
            _print_as_json_and_exit()
