# coding:utf-8

# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2022 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENSE.TXT

import sys
import time
from collections import defaultdict

from clcommon import ClPwd
from clcommon.clexception import FormattedException
from clcommon.cpapi import (Feature, NotSupported, cpinfo, domain_owner,
                            get_main_username_by_uid, getCPName, is_admin,
                            is_panel_feature_supported, is_reseller,
                            reseller_users)
from clcommon.cpapi.cpapiexceptions import EncodingError
from clcommon.lib import MySQLGovernor, MySQLGovException, GovernorStatus
from cli_utils import print_dictionary, replace_params
from cllicense import CloudlinuxLicenseLib
from cllimits import CageFs, CageFsException, LveCtl, LvectlException
from cllimits.clquota_lib import (ClQuotaException, ClQuotaLib,
                                  QuotaDisabledException)
from lvectllib import convert_to_kernel_format, is_active_reseller_limits, lve

from .arg_parsers import (ADMIN_ONLY_OPTIONS, AVAILABLE_KEYS, AVAILABLE_LIMITS,
                          AVAILABLE_LVE_KEYS, AVAILABLE_LVP_KEYS,
                          AVAILABLE_MYSQL_KEYS, AVAILABLE_MYSQL_KEYS_ALL,
                          AVAILABLE_QUOTA_KEYS,
                          CloudlinuxLimitsNoLveOptsParser,
                          CloudlinuxLimitsOptsParser)
from .utils import convert_mem_value_to_bytes


class DomainException(FormattedException):
    pass


OUT_LIMITS = {
    'speed': lambda limits, value:
    limits['cpu'].update({
        'all': str(int(convert_to_kernel_format(f'{value}%' if value.isdigit()
                                                else value)) // 100)
    }),
    'io': lambda limits, value: limits['io'].update({'all': value}),
    'mysql-cpu': lambda limits, value: limits['cpu'].update({'mysql': value}),
    'mysql-io': lambda limits, value: limits['io'].update({'mysql': value}),
    'inodes': lambda limits, value: limits['inodes'].update({
        'soft': value.split(',')[0].strip(),
        'hard': value.split(',')[1].strip()
    }),
    'maxEntryProcs': lambda limits, value: limits.update({'ep': value}),
    'nproc': lambda limits, value: limits.update({'nproc': value}),
    'pmem': lambda limits, value:
    limits.update({'pmem': convert_mem_value_to_bytes(value)}),
    'vmem': lambda limits, value:
    limits.update({'vmem': convert_mem_value_to_bytes(value)}),
    'iops': lambda limits, value: limits.update({'iops': value}),
}

ADMIN_RESELLER_NAMES = ['root', 'admin', '-']
MYSQL_GOV_ENABLED_MODE = ['abusers', 'single']


class CloudlinuxLimits:
    """
    Class for run cloudlinux-limits commands
    """
    # CLI options
    _opts = None
    # need JSON output or no
    _is_json = False
    # List of panel uid. Cache data for reuse in child classes
    _panel_uids_list = None
    # Need for GET command for store all users
    _users_list = None
    # User => reseller dictionary
    _user_to_reseller = None
    _userdomain_cache = None  # type: dict[str, str]

    def __new__(cls):
        if not is_panel_feature_supported(Feature.LVE):
            return super(CloudlinuxLimits, CloudlinuxLimitsNoLve).__new__(CloudlinuxLimitsNoLve)
        return super().__new__(cls)

    def __init__(self):
        """
        Empty constructor
        """
        # create libraries instances
        self._clpwd = ClPwd()
        self._lvectl = LveCtl()
        self._cagefs = CageFs()
        self._cl_quota = ClQuotaLib()
        self._mysql_governor = MySQLGovernor()
        self._opts_parser = CloudlinuxLimitsOptsParser()

        # set empty users list for full limits display
        self._users_list = {}

    def print_data(self, data, force_json=False, result="success"):
        """
        Output data wrapper
        :param: `dict` data - data for output to stdout
        :param: `bool` force_json - always output json format
        """
        if isinstance(data, dict):
            data.update({"result": result, "timestamp": time.time()})

        print_dictionary(data, self._is_json or force_json)

    def run(self, argv):
        """
        Run command action
        """
        licence = CloudlinuxLicenseLib()
        if not licence.get_license_status():
            self._is_json = True
            return self._error_and_exit({"result": "Cloudlinux license isn't valid"})
        self._is_json = "--json" in argv

        # parse CLI arguments
        self._opts = self._parse_args(argv)
        # If panel DA or Plesk - reseller functions are no supported
        if self._opts.get('--for-reseller') or self._opts.get('--reseller-name'):
            if not lve.is_panel_supported():
                self._error_and_exit({
                    'result': f'--for-reseller/--reseller-name options not supported on {getCPName()}.'
                })
        if self._opts.get('--for-reseller') is not None and not lve.reseller_limit_supported():
            self._error_and_exit({
                'result': 'Reseller limits are not supported. Can not use --for-reseller option.'
            })
        # Check reseller existance
        self._verify_reseller()
        self._check_user_owner()

        if self._opts["set"]:
            return self._set_run()
        if self._opts["get"]:
            return self._get_run()
        if self._opts["disable-reseller-limits"]:
            return self._disable_reseller_limits()
        if self._opts["enable-reseller-limits"]:
            return self._enable_reseller_limits()
        return self._error_and_exit({"result": "Unknown action, exit"})

    def _parse_args(self, argv):
        """
        Parse CLI arguments
        """
        status, data = self._opts_parser.parse_args(argv, self._is_json)
        if not status:
            # exit with error if can`t parse CLI arguments
            self._error_and_exit(replace_params(data))

        # change 'default' lve-id to 0
        if isinstance(data["--lve-id"], str):
            data["--lve-id"] = 0
        return data

    def _check_user_owner(self):
        """
        Checks that specified reseller is an owner of specified user and
        terminate program if need
        :return:
        """
        for_reseller = self._opts.get('--for-reseller')
        if not for_reseller:
            # No reseller specified - skip check
            return

        if self._opts.get('--lve-id') == 0:
            # Special case (default limits for reseller users) - skip check
            return

        # check user owner
        _, username = self._check_user()
        if username in [None, 'N/A']:
            # User not found - nothing to check
            # This case is totally valid(see _check_user() doc string)
            return

        if for_reseller != self._get_reseller_name(username):
            # user is not owned by specified reseller
            self._error_and_exit({
                "result": "%(username)s is not owned by reseller %(reseller)s",
                "context": {
                    "username": username,
                    "reseller": for_reseller,
                }
            })

    def _get_run(self):
        """
        Run `get` command
        """
        if not self._opts['--reseller-name'] and self._opts["--for-reseller"]:
            self._opts['--reseller-name'] = self._opts["--for-reseller"]
        # Build need users list
        self._get_check_user()

        # Get users data
        return self._get_action_list()

    def _disable_reseller_limits(self):
        """
        Run `disable-reseller-limits` command
        """
        res_id = self._opts['--reseller-name'] or self._opts['--reseller-id']
        try:
            self._lvectl.disable_reseller_limits(res_id)
            self.print_data({}, True)
        except LvectlException as e:
            self._error_and_exit({"result": e.message, "context": e.context})

    def _enable_reseller_limits(self):
        res_id = (
            self._opts['--reseller-name']
            or self._opts['--reseller-id']
            or '--all'
        )
        try:
            self._lvectl._set_container_limits_by_id_or_name(str(res_id))
            self.print_data({}, True)
        except LvectlException as e:
            self._error_and_exit({"result": e.message, "context": e.context})

    def _get_check_user(self):
        """
        Determines users according to --lve-id/--username/--domain arguments
        Or build all panel users list
        """
        # try to determine user by --lve-id/--username arguments
        uid, username = self._check_user()
        if uid is not None and username is not None:
            return uid, username

        o = self._opts
        if o["--domain"]:
            try:
                # Get user name by domain
                username = self.get_user_name_by_domain(o["--domain"])
            except DomainException as e:
                self._error_and_exit({"result": e.message, "context": e.context})

            try:
                uid = self._clpwd.get_uid(username)
            except ClPwd.NoSuchUserException as e:
                self._error_and_exit({"result": e.message, "context": e.context})

            if not self.get_package_name_by_user_id(uid) and bool(uid):
                self._error_and_exit({"result": "%(username)s is not a panel user",
                                      "context": {"username": username}})
            self._users_list[uid] = username
        else:
            # All users
            # if o["--limits"]:
            #     # Print error if user requires individual param
            #     message = "You can use options %(option)s only with single user"
            #     self._error_and_exit({'result': message, 'context': {'option': '--limits'}})
            reseller_name = o['--reseller-name']
            if reseller_name:
                resellers_users = reseller_users(reseller_name)
                for name in resellers_users:
                    try:
                        uid = self._clpwd.get_uid(name)
                    except ClPwd.NoSuchUserException as e:
                        self._error_and_exit({"result": e.message, "context": e.context})
                    self._users_list[uid] = name
                self._users_list[0] = 'root'  # default user
            else:
                try:
                    self._panel_uids_list = LveCtl.get_panel_users_uid_list()
                except EncodingError as e:
                    self._error_and_exit({"result": e.message, "context": e.context})
                name = ''
                for uid in self._panel_uids_list:
                    try:
                        name = self._clpwd.get_names(uid)[0]
                    except ClPwd.NoSuchUserException as e:
                        self._error_and_exit({"result": e.message, "context": e.context})
                    self._users_list[uid] = name
            uid = None
        return uid, username

    def get_package_name_by_user_id(self, uid):
        return self._lvectl.get_package_name_by_user_id(uid)

    def _get_username_domain_reseller(self, uid, username):
        """
        Retrives username, domain and reseller
        for supplied uid/username, for uid==0 returns '-'
        either reseller_name if has --for-reseller
        :param uid: uid
        :param username: username
        :return: dict {"id": uid, "username": username, "domain": domain, "reseller": reseller}
        """
        user = {"id": uid, "username": username, "reseller": '-'}
        if uid == 0:
            # root/default
            user["domain"] = '-'
            user["username"] = '-'
            for_reseller = self._opts.get('--for-reseller')
            user["reseller"] = for_reseller if for_reseller is not None else '-'
        else:
            # user's domain
            user_main_domain = self._get_user_main_domain(user["username"])
            if user_main_domain:
                user["domain"] = user_main_domain
            # reseller
            user['reseller'] = self._get_reseller_name(username)
            self._fix_single_uid_users(user)
        return user

    def _get_user_main_domain(self, username):
        """
        Retrives user's main domain
        :param username: User name
        """
        try:
            return self.get_user_domain(username)
        except DomainException:
            return None

    def _fix_single_uid_users(self, user: dict):
        """
        Check and fix user info if some users with same uid present. See LU-1816 for details
        :param user: User info. Example:
            {'id': 10001, 'username': 'cltest2', 'reseller': 'admin', 'domain': 'cltest2.com'}
        :return: None
        """
        uid = user['id']
        usernames = self._clpwd.get_names(uid)
        # If duplicated uid users absent - do nothing
        if len(usernames) == 1:
            return
        for username in usernames:
            main_domain = self._get_user_main_domain(username)
            reseller = self._get_reseller_name(username)
            if main_domain != 'N/A' and reseller != '-':
                user["domain"] = main_domain
                user['reseller'] = reseller
                user["username"] = username
                return

    def _get_action_list(self):
        """
        Process output for selected users
        """
        # Get all data
        result = {}
        # available limits keys
        o = self._opts

        # Check cagefs presence
        if self._cagefs.is_cagefs_present():
            result["cageFS"] = "enabled"
        else:
            result["cageFS"] = "disabled"

        if o['--for-reseller'] is None:
            # Check cl-quota presence
            if self._cl_quota.is_clquota_present():
                result["quota"] = "enabled"
            else:
                result["quota"] = "disabled"

        # Check MySQL Governor presence
        if self._mysql_governor.is_governor_present():
            result["mySqlGov"] = "enabled"
        else:
            result["mySqlGov"] = "disabled"

        lk = o["--limits"].split(",") if isinstance(o["--limits"], str) else None
        keys = dict((k, not lk or k in lk) for k in AVAILABLE_LIMITS)

        # Result users list
        users = []
        # Get limits for all users if keys `lve-id, username, domain` are None
        if o['--lve-id'] is None and o['--username'] is None and o['--domain'] is None:
            try:
                # Get information about resellers
                # Calls 'lvectl list-reseller --json --with-name'
                # and filter by reseller-name
                self._lvectl._load_info(o['--human-readable-numbers'], reseller=o['--reseller-name'])
            except LvectlException as e:
                self._error_and_exit({"result": e.message, "context": e.context})

        try:
            is_governor_available = self.can_get_governor_limits()
        except MySQLGovException as e:
            if result["mySqlGov"] == "enabled":
                result["mySqlGov"] = "error"
                if "errors" not in result:
                    result["errors"] = {}
                    result["errors"]["mySqlGov"] = str(e)
            is_governor_available = False

        # Walk through user list
        for uid, username in self._users_list.items():
            user = self._get_username_domain_reseller(uid, username)
            # Add CageFS info
            if keys["cagefs"] and self._cagefs.is_cagefs_present():
                # Cagefs is present
                try:
                    cagefs_status = self._cagefs.get_user_status(user["username"])
                    user["cageFS"] = "enabled" if cagefs_status else "disabled"
                except CageFsException as e:
                    # Retrieving data error
                    result["cageFS"] = "error"
                    if "errors" not in result:
                        result["errors"] = {}
                    result["errors"]["cageFS"] = str(e)

            user["package"] = self.get_package_name_by_user_id(uid)
            # Remove limits of previous user if keys `lve-id, domain or username` is not None
            if o['--lve-id'] is not None or o['--username'] is not None or o['--domain'] is not None:
                self._lvectl._package_data = None
            # Get limits by LVE id
            limits = self._lvectl.get_limits_by_user_id(uid, o['--human-readable-numbers'])
            if o['--for-reseller'] is None and keys["inodes"]:
                self._get_inodes_limits(uid, limits, result)

            # Add MySQL Governor limits and status
            if any([keys["mysql-gov"], keys["mysql-restrict"], keys["mysql-cpu"], keys["mysql-io"]]) \
                    and is_governor_available:
                try:
                    # Governor status
                    if keys["mysql-gov"]:
                        user["mySqlGov"] = self._mysql_governor.get_governor_status_by_username(user["username"])

                    if keys["mysql-restrict"]:
                        user["mySqlRestricted"] = self._mysql_governor.get_restrict_status_by_username(user["username"])

                    # Get governor limits
                    if keys["mysql-cpu"] or keys["mysql-io"]:
                        cpu_limit, io_limit = self._mysql_governor.get_limits_by_user(user["username"],
                                                                                      with_package_mark=True)
                        if keys["mysql-cpu"]:
                            if "cpu" not in limits:
                                limits["cpu"] = {}
                            limits["cpu"]["mysql"] = cpu_limit
                        if keys["mysql-io"]:
                            if "io" not in limits:
                                limits["io"] = {}
                            limits["io"]["mysql"] = io_limit
                except MySQLGovException as e:
                    # Retriveing data error
                    result["mySqlGov"] = "error"
                    if "errors" not in result:
                        result["errors"] = {}
                    result["errors"]["mySqlGov"] = str(e)

            # remove unneeded fields from limits
            self._filter_limits(keys, limits)
            user['limits'] = limits
            users.append(user)

        # Form output
        result["items"] = len(users)
        result["users"] = users
        if not o['--for-reseller']:
            # no --for-reseller provided - show resellers limits (if they present)
            if self._lvectl._resellers_data is not None:
                resellers_data = []
                for reseller in self._lvectl._resellers_data.values():
                    self._filter_limits(keys, reseller['limits'])
                    resellers_data.append(reseller)
                result["resellers"] = resellers_data
            else:
                result["resellers"] = []

        # print output data in JSON format
        self.print_data(result, True)
        return 0

    def _get_inodes_limits(self, uid: int, limits: dict, result: dict) -> None:
        """
        Retrieve inodes limits for user

        :param int uid: user id
        :param dict limits: limits dict
        :param dict result: result dict
        """
        if not self._cl_quota.is_clquota_present():
            result["quota"] = "error"
            result.setdefault("errors", {})["quota"] = "LVE Manager/quota package not installed"
            return

        if not self._cl_quota.is_clquota_activated():
            result["quota"] = "error"
            result.setdefault("errors", {})["quota"] = "Quotas not activated on this system"
            return

        try:
            soft_limit, hard_limit = self._cl_quota.get_inodes_limits_by_uid(uid)
            limits['inodes'] = {"soft": soft_limit, "hard": hard_limit}
        except QuotaDisabledException as e:
            result["quota"] = "error"
            result.setdefault("errors", {})["quota"] = str(e)
        except ClQuotaException as e:
            result |= {"warning": e.message, "context": e.context}

    def can_get_governor_limits(self):
        """
        Checks if the Governor limits can be retreived.

        After making sure that mysqlgov is installed and operational,
        checks that its mode is in [abusers single], user is reseller
        and resellers limits is OFF, or user belongs to admin.

        Only returns if it's True - otherwise constructs and throws MySQLGovException.
        Does it this way because these exceptions are caught in the web UI and displayed
        with optional localization.
        For localization of messages from here, see example:
        cpanel-lvemanager/spa/src/assets/i18n/hu-hu.json, key "Governor mode is not abusers or single"

        :raises MySQLGovException: Error from the Governor status check, or a constructed error
        for UI to display.
        :raises gov_error: If the Governor status check returned an error, reraises it here.
        :return: Whether it's possible to retrieve Governor limits in current state.
        :rtype: bool
        """
        if not self._mysql_governor.is_governor_present():
            exc_message = {
                'message': "%(util)s not present in system",
                'context': {'util': 'Governor'},
            }
            raise MySQLGovException(exc_message)

        gov_status, gov_error = self._mysql_governor.get_governor_status()
        if gov_status == GovernorStatus.DISABLED:
            exc_message = {
                'message': "%(util)s is currently disabled",
                'context': {'util': 'Governor'},
            }
            raise MySQLGovException(exc_message)
        if gov_status == GovernorStatus.ERROR:
            raise gov_error

        if self._mysql_governor.get_governor_mode() in MYSQL_GOV_ENABLED_MODE:
            return True

        exc_message = {
            'message': "%(util)s mode is not %(mode)s",
            'context': {'util': 'Governor', 'mode': 'abusers or single'},
        }
        raise MySQLGovException(exc_message)

    def _filter_limits(self, keys, limits):
        """
        Remove limits that are stored as key: False in `keys` dict;
        Used for --limits key;
        :type keys: dict[str, bool]
        :type limits: dict[str, str | dict]
        """
        for k in ['pmem', 'vmem', 'nproc', 'iops']:
            if not keys[k] and limits.get(k, False):
                limits.pop(k)
        if not keys['maxEntryProcs'] and limits.get('ep', False):
            limits.pop('ep')
        if not keys['speed'] and limits.get('cpu', False) and limits['cpu'].get('all', False):
            limits['cpu'].pop('all')
            if not limits['cpu']:
                limits.pop('cpu')
        if not keys['io'] and limits.get('io', False) and limits['io'].get('all', False):
            limits['io'].pop('all')
            if not limits['io']:
                limits.pop('io')

    def _set_run(self):
        """
        Run `set` command
        """
        # validate user data
        uid, username = self._set_check_user()

        self._verify_reseller()

        return self._set_action(uid, username)

    def _set_check_user(self):
        """
        Find user for modify
        """
        uid, username = self._check_user()

        opts_without_user = any(self._opts[x] for x in ["--mysql-unrestrict-all", "--reseller-name"])
        if (uid is None or username is None) and not opts_without_user:
            self._error_and_exit({'result': "You must set lve-id or username for modify limits"})

        # Don't check resellers activity for
        #  --mysql-restrict/--mysql-unrestrict-all/--mysql-gov/--mysql-io/--mysql-cpu/--cagefs/--inodes
        #  options
        # but check for --speed/--nproc/--pmem/--vmem/--maxEntryProcs/--io/--iops
        if any(self._opts[x] for x in ADMIN_ONLY_OPTIONS) \
           and not any(self._opts[f"--{x}"] for x in AVAILABLE_LVE_KEYS):
            return uid, username
        if not self._opts['--reseller-name']:
            if self._opts['--for-reseller'] is not None and not is_admin(self._opts['--for-reseller']):
                if not is_active_reseller_limits(self._opts['--for-reseller']):
                    self._error_and_exit({
                        'result': "Reseller limits for %(reseller)s are disabled.",
                        'context': {'reseller': self._opts['--for-reseller']}
                    })
            else:
                reseller_name = self._get_reseller_name(username)
                if is_active_reseller_limits(reseller_name):
                    self._error_and_exit({
                        'result': "Reseller limits for %(reseller)s are enabled. "
                                  "Admin can`t set limits for resellers users.",
                        'context': {'reseller': reseller_name}
                    })

        return uid, username

    @staticmethod
    def _clean_defaults(defaults, t):
        """
        Return keys to reset that are only acceptable to this type of limits
        :param list defaults: List of keys to reset
        :param str t: one of "lve", "cl-quota", "dbgov"
        """
        all_key = set(["all"])
        keys = {
            "lve": AVAILABLE_LVE_KEYS | all_key,
            "lvp": AVAILABLE_LVP_KEYS | all_key,
            "cl-quota": AVAILABLE_QUOTA_KEYS | all_key,
            "dbgov": AVAILABLE_MYSQL_KEYS | all_key
        }[t]
        if len(set(defaults) & keys) > 0:
            return list(set(defaults) & keys)
        return None

    def _verify_reseller(self):
        """
        If any reseller is specified in cli options ensures that it exists
        and really a reseller. Will exit with error if not.
        """
        o = self._opts
        reseller = o['--reseller-name'] or o['--for-reseller']
        if not reseller:
            return      # No reseller in options
        if not is_reseller(reseller):
            self._error_and_exit({
                'result': "No such reseller: %(reseller)s",
                'context': {'reseller': reseller}
            })

    def _set_action(self, uid, username):
        """
        Process set limits command
        :param uid: User uid
        :param username: User name
        """
        o = self._opts
        action_with_reseller_container = False
        set_default_resellers_limits = False
        reseller_limits_supported = is_panel_feature_supported(Feature.RESELLER_LIMITS)
        if reseller_limits_supported and o['--for-reseller'] and o['--lve-id'] == 0:
            # --for-reseller <res> --lve-id=0 -- Set default limits for reseller's users
            # calls lvectl set-reseller-default if self._opts['--lve-id'] is not None -- set default
            #       lvectl set-reseller -- else - self._opts['--lve-id'] == None
            is_lvectl_error, ret_list, result = self._set_reseller_action(o, uid)
            set_default_resellers_limits = True
        else:
            action_with_reseller_container = o['--reseller-name'] is not None
            o['--lve-id'] = None
            if reseller_limits_supported and action_with_reseller_container:
                # lvectl set-reseller
                is_lvectl_error, ret_list, result = self._set_reseller_action(o, uid)
            else:
                is_lvectl_error, ret_list, result = self._set_user_action(o, uid, username)

        # Never occur, due to docopt argument control.
        ret_list = ret_list or [0]
        if any(ret_list):
            # Return first error code or 42 for empty list(should never happen):
            return next((code for code in ret_list if code), 42)

        if is_panel_feature_supported(Feature.LVE) and uid == 0:
            # Do "lvectl apply all" whenever default limits has changed
            # (no matter server-wide or some reseller)
            self._lvectl.apply_all_limits()

        return self._make_set_action_response(action_with_reseller_container,
                                              is_lvectl_error, o, result,
                                              set_default_resellers_limits,
                                              uid, username)

    def _make_set_action_response(self, action_with_reseller_container,
                                  is_lvectl_error, o, result,
                                  set_default_resellers_limits, uid, username):
        # Print exit status
        if not self._is_json:
            print("success")
            return 0

        result.update({
            'timestamp': time.time(),
            'result': 'success',
        })
        if (o['--default'] or o['--unlimited']) and not is_lvectl_error:
            self._cl_quota._inode_user_limits = None
            if is_panel_feature_supported(Feature.LVE):
                self._lvectl._package_data = None
            self._get_action_list()
        else:
            is_lve_limits = False
            is_limits = False
            setted_limits = defaultdict(dict)
            if o['--mysql-cpu'] is not None or o['--mysql-io'] is not None:
                # if MySQL limits was updated, reread them from governor.
                # they will receive asterisks, if limits is different from package limits
                o['--mysql-cpu'], o['--mysql-io'] = self._mysql_governor.get_limits_by_user(username,
                                                                                            with_package_mark=True)
            for key, value in o.items():
                key = key[2:]
                if value and key in AVAILABLE_KEYS:
                    is_limits = True
                    if value and key in AVAILABLE_LVE_KEYS:
                        is_lve_limits = True
                    value = str(value)
                    value_func = OUT_LIMITS[key]
                    value_func(setted_limits, value)
            # update limits if those are lve-limits
            # they will receive asterisks, if limits is different from package limits
            if is_limits:
                if is_lve_limits:
                    self._lvectl._package_data = None
                    if set_default_resellers_limits:
                        # default limits for reseller was set
                        # get setted limits
                        getted_limits = self._lvectl.get_default_limits_by_reseller_name(o['--for-reseller'])
                    else:
                        if action_with_reseller_container:
                            # lvectl list-reseller <reseller_name> --json --with-name
                            # getted_reseller_data = dict (name:, id:, limits:, )
                            getted_reseller_data = self._lvectl.get_limits_by_reseller_name(o['--reseller-name'])
                            username = getted_reseller_data['name']
                            if uid is None:
                                uid = getted_reseller_data['id']
                            getted_limits = getted_reseller_data['limits']
                        else:
                            getted_limits = self._lvectl.get_limits_by_user_id(uid)

                    def update_func(old_value, new_value):
                        if isinstance(old_value, dict):
                            old_value.update(new_value)
                        else:
                            old_value = new_value
                        return old_value

                    setted_limits = {
                        key: update_func(value, getted_limits[key])
                        if key in getted_limits else value
                        for key, value in setted_limits.items()
                    }

                reseller_name = self._get_reseller_name(username)
                if action_with_reseller_container:
                    resellers = [{
                        'id': uid,
                        'name': username,
                        'limits': setted_limits
                    }]
                    result.update({'resellers': resellers})
                else:
                    users = [
                        {
                            'id': uid,
                            'username': '-' if username == 'root' else username,
                            'reseller': reseller_name,
                            'limits': setted_limits
                        }
                    ]
                    result.update({'users': users})
            self.print_data(result, True)
        return 0

    def _set_user_action(self, o, uid, username):
        is_lvectl_error = False
        ret_list = []
        result = {}

        reseller_name = o['--for-reseller']

        if is_panel_feature_supported(Feature.LVE) and any(o[f"--{s}"] for s in AVAILABLE_LVE_KEYS):
            # set LVE limits
            limits_map = {"maxEntryProcs": "mep"}
            # also rename maxEntryProcs or ep argument as "mep"
            limits = dict((limits_map.get(x, x), o[f"--{x}"]) for x
                          in AVAILABLE_LVE_KEYS if o[f"--{x}"] is not None)
            limits["save-all-parameters"] = o.get("--save-all-parameters")

            try:
                ret_code = self._lvectl.set_lve_limits_by_user_id(uid, limits, reseller_name)
                ret_list.append(ret_code)
            except LvectlException as e:
                self._error_and_exit({"result": e.message, "context": e.context})

        if o["--unlimited"]:
            # set unlimited limits
            if is_panel_feature_supported(Feature.LVE):
                try:
                    # set LVE limits unlimited
                    self._lvectl.set_lve_unlimited(uid, reseller_name)
                    lvectl_error_dict = None
                except (LvectlException,) as e:
                    is_lvectl_error = True
                    lvectl_error_dict = {"result": e.message, "context": e.context}

                if is_lvectl_error:
                    # lvectl error
                    result["lvectl"] = "error"
                    result["errors"] = {}
                    result["errors"]["lvectl"] = lvectl_error_dict

            try:
                # XXX: set inodes unlimited
                if reseller_name is None:
                    self._cl_quota.set_user_inodes_limits_unlimited(uid)
                is_quota_error = False
                quota_error_dict = None
            except (ClQuotaException,) as e:
                is_quota_error = True
                quota_error_dict = {"result": e.message, "context": e.context}
            if is_quota_error:
                result["quota"] = "error"
                if "errors" not in result:
                    result["errors"] = {}
                result["errors"]["quota"] = quota_error_dict

            # Successfull end
            ret_code = 0
            ret_list.append(ret_code)

        if o["--default"] is not None:
            # set uid default lve limits
            if is_panel_feature_supported(Feature.LVE):
                try:
                    # set lvectl default
                    lve_defaults = self._clean_defaults(o["--default"], "lve")
                    if lve_defaults is not None:
                        self._lvectl.reset_lve_limits(uid, lve_defaults)
                    is_lvectl_error = False
                    lvectl_error_dict = None
                except (LvectlException,) as e:
                    is_lvectl_error = True
                    lvectl_error_dict = {"result": e.message, "context": e.context}
            if is_lvectl_error:
                # lvectl error
                result["lvectl"] = "error"
                result["errors"] = {}
                result["errors"]["lvectl"] = lvectl_error_dict
            if o['--for-reseller'] is None:
                try:
                    # XXX: set inodes default
                    if self._clean_defaults(o["--default"], "cl-quota") is not None:
                        self._cl_quota.reset_user_inodes_limits(uid)
                    is_quota_error = False
                    quota_error_dict = None
                except (ClQuotaException,) as e:
                    is_quota_error = True
                    quota_error_dict = {"result": e.message, "context": e.context}
                if is_quota_error:
                    result["quota"] = "error"
                    if "errors" not in result:
                        result["errors"] = {}
                    result["errors"]["quota"] = quota_error_dict
                ######################
                # MySQL Governor
                if is_panel_feature_supported(Feature.GOVERNOR):
                    try:
                        gov_limits = self._clean_defaults(o["--default"], "dbgov")
                        if self._mysql_governor.is_governor_present() and gov_limits is not None:
                            self._mysql_governor.reset_user_limits_to_defaults(username, gov_limits)
                    except (MySQLGovException, ) as e:
                        if "errors" not in result:
                            result["errors"] = {}
                        result["errors"]["mySqlGov"] = {"result": e.message, "context": e.context}

            # Successfull end
            ret_code = 0
            ret_list.append(ret_code)

        if o["--cagefs"]:
            try:
                new_user_status = o["--cagefs"] == "enabled"
                self._cagefs.set_user_status(username, new_user_status)
                ret_code = 0
                ret_list.append(ret_code)
            except CageFsException as e:
                self._error_and_exit({"result": e.message, "context": e.context})
        if is_panel_feature_supported(Feature.GOVERNOR) and any(o[f"--{s}"] for s in AVAILABLE_MYSQL_KEYS_ALL):
            try:
                if self.can_get_governor_limits():
                    if o["--mysql-gov"]:
                        st = o["--mysql-gov"] == "watched"
                        try:
                            # set_governor_status_for_user always returns 0 or throws exception
                            ret_code = self._mysql_governor.set_governor_status_for_user(username, st)
                            ret_list.append(ret_code)
                        except MySQLGovException as e:
                            self._error_and_exit({"result": e.message, "context": e.context})

                    if o["--mysql-restrict"]:
                        st = o["--mysql-restrict"] == "restricted"
                        try:
                            ret_code = self._mysql_governor.set_restricted_status_for_user(username, st)
                            ret_list.append(ret_code)
                        except MySQLGovException as e:
                            self._error_and_exit({"result": e.message, "context": e.context})

                    if o["--mysql-unrestrict-all"]:
                        try:
                            ret_code = self._mysql_governor.set_unrestricted_status_for_all_users()
                            ret_list.append(ret_code)
                        except MySQLGovException as e:
                            self._error_and_exit({"result": e.message, "context": e.context})

                    if any(o[f"--{s}"] for s in AVAILABLE_MYSQL_KEYS):
                        cpu, io = o["--mysql-cpu"], o["--mysql-io"]

                        # LU-402. This is an ugly workaround for old Governor that don't
                        # support setting limits in bytes. When 1.2-18+ will be in
                        # production we should revert this patch and add conflict to SPEC
                        if io == 1 and not self._mysql_governor._is_governor_newer_then('1.2-17'):
                            io = 2
                            o["--mysql-io"] = 2
                        try:
                            ret_code = self._mysql_governor.set_limits_for_user(
                                'default' if username == 'root' else username, cpu, io
                            )
                            ret_list.append(ret_code)
                        except MySQLGovException as e:
                            self._error_and_exit({"result": e.message, "context": e.context})
            except MySQLGovException as e:
                self._error_and_exit({"result": e.message, "context": e.context})

        if o["--inodes"]:
            try:
                self._cl_quota.set_user_inodes_limits(
                    uid, o["--inodes"], force=o.get("--save-all-parameters", False)
                )
                ret_code = 0
                ret_list.append(ret_code)
            except ClQuotaException as e:
                self._error_and_exit({"result": e.message, "context": e.context})

        return is_lvectl_error, ret_list, result

    def _set_reseller_action(self, o, uid):
        is_lvectl_error = False
        result = {}
        ret_list = []

        if any(self._opts[f"--{s}"] for s in AVAILABLE_LVP_KEYS):
            # set LVP limits
            limits_map = {"maxEntryProcs": "mep"}
            # also rename maxEntryProcs or ep argument as "mep"
            limits = dict((limits_map.get(x, x), self._opts[f"--{x}"]) for x
                          in AVAILABLE_LVP_KEYS if self._opts[f"--{x}"] is not None)
            try:
                if self._opts['--lve-id'] is not None:
                    # Set default limits for reseller's users
                    # lvectl set-reseller-default
                    ret_code = self._lvectl.set_lvp_defaults_by_reseller_id(self._opts['--for-reseller'], limits)
                else:
                    # Set limits for LVP
                    # lvectl set-reseller
                    ret_code = self._lvectl.set_lvp_limits_by_reseller_id(self._opts['--reseller-name'], limits)
                ret_list.append(ret_code)
            except LvectlException as e:
                self._error_and_exit({"result": e.message, "context": e.context})

        if o["--default"] is not None:
            try:
                lve_defaults = self._clean_defaults(o['--default'], "lvp")
                if lve_defaults is not None:
                    # lvectl set-reseller <uid> --default=xxx
                    ret_code = self._lvectl.reset_reseller_limits(o['--reseller-name'], lve_defaults)
                    ret_list.append(ret_code)
                is_lvectl_error = False
                lvectl_error_dict = None
            except (LvectlException,) as e:
                is_lvectl_error = True
                lvectl_error_dict = {"result": e.message, "context": e.context}

            if is_lvectl_error:
                result["lvectl"] = "error"
                result["errors"] = {}
                result["errors"]["lvectl"] = lvectl_error_dict

        if o["--unlimited"]:
            try:
                self._lvectl.set_lvp_unlimited(o['--reseller-name'])
                lvectl_error_dict = None
            except (LvectlException,) as e:
                is_lvectl_error = True
                lvectl_error_dict = {"result": e.message, "context": e.context}
            if is_lvectl_error:
                result["lvectl"] = "error"
                result["errors"] = {}
                result["errors"]["lvectl"] = lvectl_error_dict
            ret_code = 0
            ret_list.append(ret_code)
        return is_lvectl_error, ret_list, result

    def _init_cache(self):
        info = cpinfo(keyls=('cplogin', 'reseller', 'dns'))
        self._user_to_reseller = dict(((i[0], i[1]) for i in info))
        self._userdomain_cache = dict(((i[0], i[2]) for i in info))

    def _get_reseller_name(self, username):
        """
        Try to get reseller for given username;
        :type username: str
        :rtype: str
        """
        if self._user_to_reseller is None:
            try:
                self._init_cache()
            except NotSupported:
                self._user_to_reseller = {}
        reseller_name = self._user_to_reseller.get(username, '-')
        return reseller_name

    def get_user_domain(self, username):
        """
        Retrive main domain for supplied user
        :param username: User name
        :return: Domain name or N/A
        """
        if self._userdomain_cache is None:
            try:
                self._init_cache()
            except NotSupported:
                self._userdomain_cache = {}
        return self._userdomain_cache.get(username, 'N/A')

    def get_user_name_by_domain(self, domain_name):
        """
        Retrive username by domain
        :param domain_name: Domain name
        :return: Username
        """
        try:
            username = domain_owner(domain_name)
            if username:
                return username
        except (IOError, NotSupported) as e:
            raise DomainException(str(e)) from e
        raise DomainException({'message': "There is no domain '%(domain)s'",
                               'context': {'domain': domain_name}})

    def _check_user(self):
        """
        User identification from CLI options. Common for all commands.
        May return (None, None) when user wasn't specified in any way and this
        is a *valid* case when we want to list all users with command like:
        "... get --json --for-reseller=reseller1"
        """
        o = self._opts
        uid, username = None, None

        def _username_message(username):
            return {'result': '%(username)s is not a panel user', 'context': {'username': username}}

        if o["--lve-id"] or o["--lve-id"] == 0:
            uid = o["--lve-id"]
            user_package = self.get_package_name_by_user_id(uid)
            username = get_main_username_by_uid(uid)

            # We don't work with non-panel users
            if not user_package and bool(uid):

                def _uid_message(uid):
                    return {'result': '%(uid)s is not a panel user', 'context': {'uid': uid}}

                self._error_and_exit(_uid_message(uid))

            self._users_list[uid] = username
        elif o["--username"]:
            username = o["--username"]
            try:
                # Get id by username
                uid = self._clpwd.get_uid(username)
            except ClPwd.NoSuchUserException as e:
                self._error_and_exit({"result": e.message, "context": e.context})
            user_package = self.get_package_name_by_user_id(uid)
            if not user_package and bool(uid):
                self._error_and_exit(_username_message(username))

            self._users_list[uid] = username

        return uid, username

    def _error_and_exit(self, message, error_code=1):
        """
        Print error and exit

        :param message dict: Dictionary with keys "result" as string and optional "context" as dict
        """
        if self._is_json:
            message.update({"timestamp": time.time()})
            print_dictionary(message, True)
        else:
            try:
                print(str(message["result"]) % message.get("context", {}))
            except KeyError as e:
                print(f"Error: {e} [{message}]")
        sys.exit(error_code)


class CloudlinuxLimitsNoLve(CloudlinuxLimits):

    def __init__(self):  # pylint: disable=super-init-not-called
        self._clpwd = ClPwd()
        self._cagefs = CageFs()
        self._cl_quota = ClQuotaLib()
        self._opts_parser = CloudlinuxLimitsNoLveOptsParser()
        self._users_list = {}

    def get_package_name_by_user_id(self, uid):
        """
        Get package name by user id.
        None package returns as str(None) for user checker compatibility:
          'not str(None)' is True
        :param: id - user/lve id
        :return: Package name
        """
        from clcommon.cpapi import reseller_package_by_uid  # pylint: disable=import-outside-toplevel
        try:
            package_name = str(reseller_package_by_uid(uid)[1])
        except IndexError:
            # No user with such uid, use empty package name
            package_name = ''
        return package_name

    def _get_action_list(self):
        """
        Process output for selected users
        """
        # Get all data
        result = {}
        # available limits keys
        o = self._opts

        # Check cagefs presence
        if self._cagefs.is_cagefs_present():
            result["cageFS"] = "enabled"
        else:
            result["cageFS"] = "disabled"

        if o['--for-reseller'] is None:
            # Check cl-quota presence
            if self._cl_quota.is_clquota_present():
                result["quota"] = "enabled"
            else:
                result["quota"] = "disabled"

        lk = o["--limits"].split(",") if isinstance(o["--limits"], str) else None
        keys = dict((k, not lk or k in lk) for k in AVAILABLE_LIMITS)

        # Result users list
        users = []
        # Walk throw user list
        for uid, username in self._users_list.items():
            user = self._get_username_domain_reseller(uid, username)
            # Add CageFS info
            if keys["cagefs"] and self._cagefs.is_cagefs_present():
                # Cagefs is present
                try:
                    cagefs_status = self._cagefs.get_user_status(user["username"])
                    user["cageFS"] = "enabled" if cagefs_status else "disabled"
                except CageFsException as e:
                    # Retriveing data error
                    result["cageFS"] = "error"
                    if "errors" not in result:
                        result["errors"] = {}
                    result["errors"]["cageFS"] = str(e)

            user["package"] = self.get_package_name_by_user_id(uid)
            limits = {}
            if o['--for-reseller'] is None and keys["inodes"]:
                self._get_inodes_limits(uid, limits, result)
            user['limits'] = limits
            users.append(user)

        # Form output
        result["items"] = len(users)
        result["users"] = users
        if not o['--for-reseller']:
            result["resellers"] = []

        # print output data in JSON format
        self.print_data(result, True)
        return 0
