# -*- coding: utf-8 -*-

# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2022 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENSE.TXT

# packages.py - work code for cloudlinux-packages utility

import sys
import time

from clcommon.const import Feature
from clcommon.cpapi import admin_packages, is_admin, is_panel_feature_supported, resellers_packages
from clcommon.lib import MySQLGovernor, MySQLGovException
from cli_utils import print_dictionary, replace_params
from cllicense import CloudlinuxLicenseLib
from cllimits import LvectlException
from cllimits.clquota_lib import ClQuotaException, ClQuotaLib, QuotaDisabledException
from clpackages import LveCtlPackage
from lvectllib import DEFAULT_PACKAGE, is_active_reseller_limits, lve

from .cl_packages_arg_parse import parse_cloudlinux_packages_opts


class CloudlinuxPackages:
    """
    Class for run cloudlinux-packages commands
    """
    _LIMITS_KEYS = ['speed', 'nproc', 'pmem', 'vmem', 'maxEntryProcs', 'io', 'iops', 'inodes', 'mysql-cpu', 'mysql-io']

    def __init__(self):
        self._is_json = False
        self._opts = {}
        self._clquota_pack_lib = ClQuotaLib()
        self._lvectl_pack_lib = LveCtlPackage() if is_panel_feature_supported(Feature.LVE) else None
        self._mysql_governor = MySQLGovernor() if is_panel_feature_supported(Feature.GOVERNOR) else None

    def run(self, argv):
        """
        Run command action
        """
        licence = CloudlinuxLicenseLib()
        if not licence.get_license_status():
            self._is_json = True
            return self._error_and_exit({"result": "Cloudlinux license isn't valid"})
        self._is_json = "--json" in argv
        # get arguments
        self._opts = self._parse_args(argv)
        if self._opts.get('--for-reseller') and not lve.reseller_limit_supported():
            self._error_and_exit({
                'result': "error: reseller limits are not supported, can not use --for-reseller option."
            })

        if self._opts["set"]:
            return self._set_run()
        return self._get_run()

    def _parse_args(self, argv):
        """
        Parse CLI arguments
        """
        status, data = parse_cloudlinux_packages_opts(argv, self._is_json)
        if not status:
            # exit with error if can`t parse CLI arguments
            self._error_and_exit(replace_params(data))
        return data

    def _get_run(self):
        """
        Run 'get' command
        """
        # available limits keys
        o = self._opts
        lk = o["--limits"].split(",") if isinstance(o["--limits"], str) else None
        limits_keys = dict((k, not lk or k in lk) for k in self._LIMITS_KEYS)
        result = {}
        reseller_name = self._opts['--provider']
        package_name = None
        if self._opts['--package'] is not None:
            # TODO: py3 we can remove this unicode-escape transformation
            package_name = self._opts['--package'].encode().decode('unicode-escape')

        packages_data: list[dict] = []
        if is_panel_feature_supported(Feature.LVE):
            try:
                packages_data = self._lvectl_pack_lib.get_reseller_package_limits(reseller_name,
                                                                                package_name,
                                                                                self._opts['--human-readable-numbers'])
            except LvectlException as e:
                self._error_and_exit({"result": e.message, "context": e.context})
        # Add quota limits if they are supported and need
        if limits_keys['inodes']:
            if not self._clquota_pack_lib.is_clquota_present():
                # Quotas not present
                result["quota"] = "error"
                result["errors"] = {}
                result["errors"]["quota"] = "LVE Manager/quota package not installed"
            elif not self._clquota_pack_lib.is_clquota_activated():
                # Quotas are not activated on system
                result["quota"] = "error"
                result["errors"] = {}
                result["errors"]["quota"] = "Quotas not activated on this system"
            else:
                # Quotas are supported, get packages limits
                try:
                    quotas_dict = self._clquota_pack_lib.get_reseller_package_limits(package_name)
                except QuotaDisabledException as e:
                    quotas_dict = {}
                    result["errors"] = {}
                    result["errors"]["quota"] = str(e)
                    result["quota"] = "error"
                except ClQuotaException as e:
                    quotas_dict = {}
                    result["warning"] = e.message
                    result["context"] = e.context
                if is_panel_feature_supported(Feature.LVE):
                    for package_data in packages_data:
                        package_name_quota = package_data['package']
                        if package_name_quota in quotas_dict:
                            package_data['limits']['inodes'] = quotas_dict.pop(package_name_quota)
                else:
                    # Try to resemble pacakge -> reseller mapping logic
                    # derived from self._clquota_pack_lib.get_reseller_package_limits
                    # and `lvectl all-package-list` implementation
                    package_to_resellers = {
                        package: reseller
                        for reseller, packages in resellers_packages().items() for package in packages
                    } | {
                        package: "root" for package in admin_packages()
                    } | {"default": ""}
                    packages_data += [
                        {
                            "package": package_name if package_name != "default" else DEFAULT_PACKAGE,
                            "provider": package_to_resellers.get(package_name, ""),
                            "enabledLimits": False,
                            "limits": {"inodes": inodes_limits}
                        }
                        for package_name, inodes_limits in quotas_dict.items()
                    ]

        # Add MySQL governor limits placeholders
        result['mySqlGov'] = 'disabled'
        if is_panel_feature_supported(Feature.LVE):
            package_names = []
            for package_data in packages_data:
                # Note: package_data['limits']['io'] -- already dict with 'all' key
                package_data['limits']['cpu'] = {'all': package_data['limits']['speed']}
                package_names.append(package_data['package'])
        if is_panel_feature_supported(Feature.GOVERNOR) and self._mysql_governor.is_governor_present():
            # MySQL Governor installed, add governor packages limits
            result['mySqlGov'] = 'enabled'
            # Add governor packages limits
            try:
                # MySQL Governor utility gives empty output if package absent in system,
                # so we need to call it only if package present in system
                if self._opts['--package'] is None or self._opts['--package'] in package_names:
                    package_limits_governor = self._mysql_governor.get_package_limits(self._opts['--package'])
                    for package_data in packages_data:
                        package_name = package_data['package']
                        # default package in MySQL Governor called 'default'
                        if package_name == DEFAULT_PACKAGE:
                            package_name = 'default'
                        if package_name in package_limits_governor:
                            package_data['limits']['cpu']['mysql'] = package_limits_governor[package_name]['cpu']
                            package_data['limits']['io']['mysql'] = package_limits_governor[package_name]['io']
            except MySQLGovException as e:
                if result["mySqlGov"] == "enabled":
                    result["mySqlGov"] = "error"
                    if "errors" not in result:
                        result["errors"] = {}
                        result["errors"]["mySqlGov"] = str(e)

        # Remove limits according to limits_keys
        for package_data in packages_data:
            limits_dict = package_data['limits']
            if not limits_keys['speed']:
                # Remove SPEED
                del limits_dict['speed']
                del limits_dict['cpu']['all']
                if limits_dict['cpu'] == {}:
                    del limits_dict['cpu']
            if not limits_keys['vmem']:
                # Remove VMEM
                del limits_dict['vmem']
            if not limits_keys['pmem'] and 'pmem' in limits_dict:
                # Remove PMEM
                del limits_dict['pmem']
            if not limits_keys['maxEntryProcs']:
                # Remove EP
                del limits_dict['ep']
            if not limits_keys['nproc'] and 'nproc' in limits_dict:
                # Remove NPROC
                del limits_dict['nproc']
            if not limits_keys['io']:
                # Remove IO
                del limits_dict['io']['all']
                if limits_dict['io'] == {}:
                    del limits_dict['io']
            if not limits_keys['iops'] and 'iops' in limits_dict:
                # Remove IOPS
                del limits_dict['iops']
            if not limits_keys['mysql-cpu'] and 'mysql' in limits_dict['cpu']:
                # Remove --mysql_cpu
                del limits_dict['cpu']['mysql']
                if limits_dict['cpu'] == {}:
                    del limits_dict['cpu']
            if not limits_keys['mysql-io'] and 'io' in limits_dict and 'mysql' in limits_dict['io']:
                # Remove --mysql-io ['io']['mysql']
                del limits_dict['io']['mysql']
                if limits_dict['io'] == {}:
                    del limits_dict['io']
        # Form output
        result['items'] = len(packages_data)
        result['package'] = packages_data
        self._print_data(result)
        return 0

    def _set_run(self):
        """
        Run 'set' command
        """
        reseller_name = self._opts['--provider']
        for_reseller = self._opts.get('--for-reseller')
        if for_reseller is not None and not is_admin(for_reseller) and \
           not is_active_reseller_limits(for_reseller):
            self._error_and_exit({
                'result': "error: reseller limits for %(reseller)s are disabled.",
                'context': {'reseller': reseller_name}
            })

        # TODO: py3 we can remove this unicode-escape transformation
        package_name = self._opts['--package'].encode().decode('unicode-escape')
        result = {}
        o = self._opts

        if is_panel_feature_supported(Feature.LVE) and any(
            [o["--speed"], o["--pmem"], o["--vmem"], o["--maxEntryProcs"], o["--nproc"], o["--io"], o["--iops"]]
        ):
            # set LVE limits
            limits = {"speed": o["--speed"], "pmem": o["--pmem"], "io": o["--io"],
                      "vmem": o["--vmem"], "maxEntryProcs": o["--maxEntryProcs"],
                      "nproc": o["--nproc"], "iops": o["--iops"]}

            # LU-505: root should not set reseller's package limits
            self._check_hoster_permissions_for_package(for_reseller, package_name, reseller_name)

            # remove None elements
            limits = {key: value for key, value in limits.items() if value is not None}
            try:
                self._lvectl_pack_lib.set_reseller_package_limits(reseller_name, package_name, limits)
            except LvectlException as e:
                self._error_and_exit({"result": e.message, "context": e.context})

        if o["--inodes"]:
            # inodes
            try:
                self._clquota_pack_lib.set_reseller_package_limits(package_name, o["--inodes"])
            except ClQuotaException as e:
                self._error_and_exit({"result": e.message, "context": e.context})

        if is_panel_feature_supported(Feature.GOVERNOR) \
                and (o["--mysql-cpu"] is not None or o["--mysql-io"] is not None) \
                and self._mysql_governor.is_governor_present():
            try:
                # set MySQL governor limits
                self._mysql_governor.set_package_limits(package_name, o["--mysql-cpu"], o["--mysql-io"])
            except (MySQLGovException, ) as e:
                self._error_and_exit({"result": e.message, "context": e.context})

        self._print_data(result)
        return 0

    def _check_hoster_permissions_for_package(self, for_reseller, package_name, reseller_name):
        # for reseller is None only when command called from admin UI
        if for_reseller is None and is_active_reseller_limits(reseller_name):
            message = ("error: hoster cannot manage package limits"
                       " for reseller's package %(package)s")
            self._error_and_exit({'result': message,
                                  'context': {'package': package_name}})

    def _print_data(self, data, force_json=False, result="success"):
        """
        Output data wrapper
        :param: `dict` data - data for output to stdout
        :param: `bool` force_json - always output json format
        """
        if isinstance(data, dict):
            data.update({"result": result, "timestamp": time.time()})
        print_dictionary(data, self._is_json or force_json)

    def _error_and_exit(self, message, error_code=1):
        """
        Print error and exit
        :param dict message: Dictionary with keys "result" as string and optional "context" as dict
        """
        if self._is_json:
            message.update({"timestamp": time.time()})
            print_dictionary(message, True)
        else:
            try:
                print(str(message["result"]) % message.get("context", {}))
            except KeyError as e:
                print(f"Error: {e} [{message}]")
        sys.exit(error_code)
